#
# Copyright 2018 Fairphone B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Android app basic representation and management."""

from pathlib import Path
import re
import subprocess
from typing import Tuple

from android_tools.utils import aapt, AaptCommandError


class App:
    """An Android app.

    Attributes:
        name: The user-facing label.
        package: The package name.
        code: The internal version.
        version: The user-facing version.

    """

    _AAPT_APPLICATION_METADATA = re.compile(
        r"application:.*label='(?P<label>[^']+)'.*"
    )
    _APPT_PACKAGE_METADATA = re.compile(
        r"package:.*"
        r"name='(?P<package>[\w.-]+)'.*"
        r"versionCode='(?P<code>\d+)'.*"
        r"versionName='(?P<version>[\w.-]+)'.*"
    )

    def __init__(self, apk: Path):
        """Initialise the app.

        The app metadata is extracted from its APK file.

        Attributes:
            apk: The path to the app package (APK).

        """
        self.apk = apk
        self.name, self.package, self.code, self.version = self.get_metadata(
            str(self.apk)
        )

    def __str__(self) -> str:
        """Return a simple representation of the app."""
        return "{} {}".format(self.name, self.version)

    def __repr__(self) -> str:
        """Return an unambiguous representation of the app."""
        return "{} ({})".format(self.package, self.code)

    def aapt(self, *args: str) -> subprocess.CompletedProcess:
        """Run an AAPT command against the device.

        Arguments:
            *args: The AAPT command with its options.
        Returns:
            The completed process.
        Raises:
            AaptCommandError: If the command failed (i.e. return code < 0).

        """
        try:
            return aapt(*args, apk=self.apk)
        except AaptCommandError as error:
            if error.target == str(self.apk):
                # Raise more useful information than just the device serial
                raise AaptCommandError(
                    repr(self), error.command, error.error_message
                )
            else:
                raise error

    @classmethod
    def get_metadata(cls, apk: Path) -> Tuple[str, str, str, str]:
        """Read an app metadata from an APK.

        Arguments:
            apk: The path to the APK.
        Returns:
            A tuple composed of the application label (user-facing name), the
            package name, the version code, and the version name.
        Raises:
            AaptCommandError: If the AAPT command failed.
            ValueError: If not exactly one package was found in the APK.

        """
        label = None
        package = None
        code = None
        version = None

        process = aapt("dump", "badging", apk=apk)
        for line in process.stdout.splitlines():
            match = cls._AAPT_APPLICATION_METADATA.fullmatch(line)
            if match:
                label = match.group("label")
                continue
            match = cls._APPT_PACKAGE_METADATA.fullmatch(line)
            if match:
                package = match.group("package")
                code = match.group("code")
                version = match.group("version")
                continue

        if not package or not code or not version:
            raise ValueError(
                "Could not find the package metadata in `{}`".format(apk)
            )
        if not label:
            raise ValueError(
                "Could not find the application metadata in `{}`".format(apk)
            )
        return label, package, code, version
