#
# Copyright 2018 Fairphone B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Utilities to access the host Android SDK Platform Tools."""

from pathlib import Path
import subprocess

from adb.adb_protocol import AuthSigner


_ANDROID_KEYS_FROM_USER_HOME = Path.home() / ".android" / "adbkey"


def auth_signer_from_user_keys() -> AuthSigner:
    """Create a new AuthSigner from the user (Android) keys.

    Create an `AuthSigner` from the RSA keys located in the user home directory
    as per the Android SDK usage. By default, keys are stored as
    `~/.android/adbkey[.pub]`.

    Returns:
        A new `AuthSigner` based on the user's (Android) RSA keys.

    """
    from adb.sign_pythonrsa import PythonRSASigner

    return PythonRSASigner.FromRSAKeyPath(str(_ANDROID_KEYS_FROM_USER_HOME))


class CommandError(Exception):
    """A generic representation of a failed command issued against a target."""

    def __init__(self, target: str, command: str, error_message: str):
        """Initialise the exception.

        Attributes:
            target: The command target.
            command: The command that failed.
            error_message: The error message returned by the failed command.

        """
        self.target = target
        self.command = command
        self.error_message = error_message
        super().__init__(
            "Command `{}` failed on {}: {}".format(
                self.command, self.target, self.error_message
            )
        )


class AaptCommandError(CommandError):
    """An AAPT command failed."""


class AdbCommandError(CommandError):
    """An ADB command failed."""

    def __init__(self, device: str, command: str, error_message: str):
        """Initialise the exception.

        If the target device is `None`, use the generic target "HOST" instead.
        """
        super().__init__(device if device else "HOST", command, error_message)


class FastbootCommandError(CommandError):
    """A fastboot command failed."""

    def __init__(self, device: str, command: str, error_message: str):
        """Initialise the exception.

        If the target device is `None`, use the generic target "HOST" instead.
        """
        super().__init__(device if device else "HOST", command, error_message)


def aapt(*args: str, apk: Path) -> subprocess.CompletedProcess:
    """Run an AAPT command.

    Example:
    >>> process = aapt("list", apk=pathlib.Path("CtsSampleDeviceApp.apk"))
    >>> process.returncode
    0
    >>> process.stdout.strip().splitlines()
    ['AndroidManifest.xml', 'classes.dex', 'META-INF/MANIFEST.MF',
        'META-INF/CERT.SF', 'META-INF/CERT.RSA']

    Arguments:
        *args: The AAPT command with its options.
        apk: The path to the target app package (APK).
    Returns:
        The completed process.
    Raises:
        AaptCommandError: If the command failed (i.e. return code != 0).

    """
    command = ["aapt"] + list(args) + [str(apk)]
    ret = subprocess.run(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )

    if ret.returncode != 0:
        raise AaptCommandError(str(apk), str(command), ret.stderr)

    return ret


def adb(*args: str, serial: str = None) -> subprocess.CompletedProcess:
    """Run an ADB command.

    Example:
    >>> process = adb("shell", "getprop", "ro.build.fingerprint",
        serial="567646b8")
    >>> process.returncode
    0
    >>> process.stdout.strip()
    'Fairphone/FP2/FP2:7.1.2/18.09.2/gms-4a5ca353:user/release-keys'

    Arguments:
        *args: The ADB command with its options.
        serial: The target device serial number.
    Returns:
        The completed process.
    Raises:
        AdbCommandError: If the command failed (i.e. return code != 0).

    """
    # Make sure the adb server is started to avoid the infamous
    # "out of date" message that pollutes stdout.
    command = ["adb", "start-server"]
    ret = subprocess.run(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    if ret.returncode < 0:
        raise AdbCommandError("host", str(command), ret.stderr)

    command = ["adb"]
    if serial:
        command += ["-s", serial]
    command += list(args)
    ret = subprocess.run(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )

    if ret.returncode != 0:
        raise AdbCommandError(serial, str(command), ret.stderr)

    return ret


def fastboot(*args: str, serial: str = None) -> subprocess.CompletedProcess:
    r"""Run a fastboot command.

    Example with no serial:
    >>> process = fastboot("devices")
    >>> process.returncode
    0
    >>> process.stdout.strip()
    '3a52b6bb\tfastboot'

    Example with a serial:
    >>> process = fastboot("getvar", "product", serial="3a52b6bb")
    >>> process.returncode
    0
    >>> process.stderr.strip().splitlines()
    ['product: FP2', 'finished. total time: 0.002s']

    Arguments:
        *args: The fastboot command with its options.
        serial: The target device serial number.
    Returns:
        The completed process.
    Raises:
        FastbootCommandError: If the command failed (i.e. return code != 0).

    """
    command = ["fastboot"]
    if serial:
        command += ["-s", serial]
    command += list(args)
    ret = subprocess.run(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )

    if ret.returncode != 0:
        raise FastbootCommandError(serial, str(command), ret.stderr)

    return ret
